(******************************************************************************)
(* Ce programme est libre, vous pouvez le diffuser et/ou le modifier          *)
(* en respectant les termes de la licence CeCILL                              *)
(* labore par le CEA, le CNRS et l'INRIA.                                   *)
(*                                                                            *)
(* Cette licence est trs semblable  la GPL, elle a simplement t rendue    *)
(* compatible avec le droit franais. Si vous rutilisez ce code dans un      *)
(* autre programme, votre programme peut tre distribu sous les termes de la *)
(* General Public Licence.                                                    *)
(*                                                                            *)
(* Ecrit par Jean-Baptiste Rouquier.                                          *)
(******************************************************************************)


(*Eh oui, ce programme est crit dans notre magnifique langage des prpas : Caml :*)

(*avant de compiler avec camlc, modifier la ligne de "main ()"
      |Chgt Quitter -> raise (Chgt Quitter) (*c'est ici qu'il faut mettre exit 0*)
en
      |Chgt Quitter -> exit 0

ainsi que celle de "qui_commence ()"
  if      (bas+80 <= !y) & (!y <= bas+114) then raise (Chgt Quitter) (*ici aussi, remplacer par exit 0*)
en
  if      (bas+80 <= !y) & (!y <= bas+114) then exit 0
*)

(*on identifie (type) et (argument de ce type) dans certaines fonctions*)

(******************************************************************)
(*prliminaires*)

type coord == int*int
 and coup  == coord*coord
 and case   = |Rouge|Bleu|Socle|Trou
 and joueur = |Homme|Machine
 and plateau  = {case_vide:coord; pos:case vect vect}
 and pions    = {homme:int; machine:int}
 and position = {plateau:plateau; coup_interdit:coup; pions_restants:pions}
;;

let case_of_joueur = function
|Homme   -> Bleu
|Machine -> Rouge
;;

type choix = |Niveau|Nouvelle_partie|Quitter|Indtermin;;
exception Cnv  of string;;
exception Chgt of choix;;

#open "graphics";;
open_graph "";;
let gauche= (size_x ()-300)/2 and bas= (size_y () -300)/2;;
let mine=0 and maxe=100;;

type best_coup = {bcoup:coup; valeur:int};;
exception Tangence;;

(******************************************************************)
(*entres-sorties*)

let set_couleur case = match case with
|Rouge -> set_color red
|Bleu  -> set_color blue
|Socle -> set_color black
|Trou  -> set_color white
;;

let draw_pion coord =
let (x,y)=coord in
fill_circle (gauche+100*x+50) (bas+100*y+50) 40
;;

let draw_socle coord =
let (x,y)=coord in fill_rect (gauche+100*x+5) (bas+100*y+5) 90 90
;;

let draw_case case coord =
match case with
|Rouge|Bleu -> set_color black; draw_socle coord;
	set_couleur case; draw_pion coord;
|Socle|Trou -> set_couleur case; draw_socle coord
;;

let draw_plateau plateau =
let v=plateau.pos in
for i=0 to 2 do
  for j=0 to 2 do
    draw_case v.(i).(j) (i,j);
  done;
done
;;

let encadre bool coord =
set_color (if bool then cyan else white);
let (a,b)=coord in let (x,y)=(gauche+100*a, bas+100*b) in
fill_rect x y 5 100; fill_rect (x+5) y 95 5;
fill_rect (x+5) (y+95) 95 5; fill_rect (x+95) (y+5) 5 90
;;

let mouse_coord pos_souris =
let (x,y) = pos_souris in
if (x>=gauche) & (x<=gauche+300) & (y>=bas) & (y<=bas+300)
then ( (x-gauche)/100, (y-bas)/100 )
else raise (Cnv "Cliquer sur le plateau !")
;;

let get_coup_homme () =
let dpart = ref (0,0) and arrive = ref (0,0) in
while !dpart = !arrive do
  let status=wait_next_event [Button_down;Key_pressed] in
  if status.keypressed = true then raise (Chgt Indtermin);
  let (x,y) = (status.mouse_x, status.mouse_y) in
  dpart:= mouse_coord (x,y);
  encadre true !dpart;

  let status=wait_next_event [Button_down;Key_pressed] in
  if status.keypressed = true then raise (Chgt Indtermin);
  let (x,y) = (status.mouse_x, status.mouse_y) in
  arrive:= mouse_coord (x,y);
  encadre false !dpart;
done; (!dpart,!arrive)
;;

let changement () =
set_color cyan;
  fill_rect 10 (bas+80) 121 34;
  fill_rect 10 (bas+40) 121 34;
  fill_rect 10 bas      121 34;
set_color black;
  moveto 17 (bas+140); draw_string "Cliquez  ct pour reprendre";
  moveto 17 (bas+87 ); draw_string "    Niveau     ";
  moveto 17 (bas+47 ); draw_string "Nouvelle partie";
  moveto 17 (bas+ 7 ); draw_string "    Quitter    ";
let status=wait_next_event [Button_down] in let (x,y) = (status.mouse_x,status.mouse_y) in
if (10 <= x) & (x <= 121) then
  if      (bas+80 <= y) & (y <= bas+114) then raise (Chgt Niveau)
  else if (bas+40 <= y) & (y <= bas+ 74) then raise (Chgt Nouvelle_partie)
  else if (bas    <= y) & (y <= bas+ 34) then raise (Chgt Quitter);
()
;;

let qui_commence () =
set_color cyan;
  fill_rect 10 (bas+80) 121 34;
  fill_rect 10 (bas+40) 121 34;
  fill_rect 10 bas      121 34;
set_color black;
  moveto 17 (bas+87 ); draw_string "    Quitter    ";
  moveto 17 (bas+47 ); draw_string "  Je commence  ";
  moveto 17 (bas+ 7 ); draw_string "Vous commencez ";
let x=ref 0 and y=ref 0 and result=ref true in
while not ( (10 <= !x) & (!x <= 121) & (bas <= !y) & (!y <= bas+114) ) do
  let status=wait_next_event [Button_down] in
  x:=status.mouse_x;
  y:=status.mouse_y;
  if      (bas+80 <= !y) & (!y <= bas+114) then raise (Chgt Quitter) (*ici aussi, remplacer par exit 0*)
  else if (bas+40 <= !y) & (!y <= bas+74) then result:=true
  else if (bas    <= !y) & (!y <= bas+34) then result:=false
  else x:=0;  (*pour reboucler*)
done;
!result
;;

let choisir_niveau () =
set_color cyan;
  fill_rect 10 (bas+120) 121 34;
  fill_rect 10 (bas+80 ) 121 34;
  fill_rect 10 (bas+40 ) 121 34;
  fill_rect 10 bas       121 34;
set_color black;
  moveto 17 (bas+127); draw_string "   Dbutant    ";
  moveto 17 (bas+87 ); draw_string "  Pas mauvais  ";
  moveto 17 (bas+47 ); draw_string "  Bon joueur   ";
  moveto 17 (bas+ 7 ); draw_string "Joueur patient ";
let x=ref 0 and y=ref 0 and result=ref 1 in
while not ( (10 <= !x) & (!x <= 121) & (bas <= !y) & (!y <= bas+154) ) do
  let status=wait_next_event [Button_down] in
  x:=status.mouse_x;
  y:=status.mouse_y;
  if      (bas+120 <= !y) & (!y <= bas+154) then result:=1
  else if (bas+80  <= !y) & (!y <= bas+114) then result:=2
  else if (bas+40  <= !y) & (!y <= bas+74 ) then result:=3
  else if (bas     <= !y) & (!y <= bas+34 ) then result:=4
  else x:=0;
done;
!result
;;

let print_pions compteur =
set_color black;
moveto 10 40;
  draw_string "joueur : ";
  draw_string (string_of_int compteur.homme);
moveto 10 10;
  draw_string "ordinateur : ";
  draw_string (string_of_int compteur.machine);
;;

(*fin de l'interface graphique, on passe au choses srieuses*)
(******************************************************************)


let compl i = match i with
|0 -> (1,2)
|1 -> (0,2)
|2 -> (0,1)
|_ -> failwith "compl"
;;

let coups_possibles joueur position =
let (x,y)= position.plateau.case_vide and v=position.plateau.pos
and pos_pions = ref [] and pos_socles = ref []
and couleur_joueur = case_of_joueur joueur in
for i=0 to 2 do
  for j=0 to 2 do
    let a=v.(i).(j) in
    if      a=couleur_joueur  then pos_pions := (i,j) :: !pos_pions
    else if a=Socle then pos_socles:= (i,j) :: !pos_socles
  done;
done;
let (x1,x2)=compl x and (y1,y2)=compl y in
let result=ref [((x1,y),(x,y)); ((x2,y),(x,y)); ((x,y1),(x,y)); ((x,y2),(x,y))] in
if (match joueur with Homme -> position.pions_restants.homme |Machine -> position.pions_restants.machine) >0
   then pos_pions:= (x,y) :: !pos_pions;
while !pos_pions <> [] do
  let pos_socles_copy = ref !pos_socles in
	(*let dpart::t = !pos_pions in*)
  let dpart=hd !pos_pions and t=tl !pos_pions in
  pos_pions:= t;
  while !pos_socles_copy <> [] do
    let arrive=hd !pos_socles_copy and t=tl !pos_socles_copy in
    pos_socles_copy:=t;
    result:= (dpart,arrive) :: !result;
  done;
done; except position.coup_interdit !result
;;

let pousse coup v =
let ((a,b),(c,d))= coup in
if   a=c then if abs (d-b)=1 then v.(a).(d) <- v.(a).(b)
                             else ( v.(a).(d) <- v.(a).(1); v.(a).(1) <- v.(a).(b) )
         else if abs (a-c)=1 then v.(c).(b) <- v.(a).(b)
                             else ( v.(c).(b) <- v.(1).(b); v.(1).(b) <- v.(a).(b) );
v.(a).(b) <- Trou;
;;

let copy_pos p=
let result = make_vect 3 [||] in
for i=0 to 2 do
  result.(i) <- copy_vect p.(i)
done;
result;
;;

let joue_coup coup joueur position =
let pph = ref position.pions_restants.homme and ppm = ref position.pions_restants.machine in
let ((a,b),(c,d)) = coup in
let coup_interdit_return = ref ((c,d),(a,b)) in
let v=copy_pos position.plateau.pos and cv=ref position.plateau.case_vide in
(match v.(a).(b) with
|Rouge|Bleu -> if v.(c).(d) = Trou then (pousse coup v; cv:=(a,b))
               else begin v.(a).(b) <- Socle;
                          v.(c).(d) <- case_of_joueur joueur
                    end
|Trou       -> begin v.(c).(d) <- case_of_joueur joueur;
                     if joueur=Homme then pph:= !pph -1
                                     else ppm:= !ppm -1;
                     coup_interdit_return:=((0,0),(0,0)); 
               end
|Socle      -> (pousse coup v; cv:=(a,b))
);
{plateau={case_vide= !cv; pos=v};
 coup_interdit= !coup_interdit_return;
 pions_restants={homme= !pph; machine= !ppm}}
;;

let joue_coup_homme position =
let l=coups_possibles Homme position and coup = ref ((0,0),(0,0)) in
while not (mem !coup l) do
  try coup:=get_coup_homme () with
  |Cnv "Cliquer sur le plateau !" -> (moveto 10 bas; set_color black;
          draw_string "Cliquer sur le plateau !")
  |Chgt Indtermin               -> (changement (); set_color white;
          fill_rect 0 bas (gauche-1) (bas+114) )
done;
let result = joue_coup !coup Homme position in
clear_graph (); draw_plateau result.plateau;
result;
;;

(********************************************************)
(*et ce qu'il manque pour que l'ordinateur joue :*)

let test_gagn joueur pos =
let pos_pions = ref [] in
let case=case_of_joueur joueur in
for i=0 to 2 do
  for j=0 to 2 do
    if pos.(i).(j) = case then pos_pions:= (i,j):: !pos_pions;
  done;
done;
if (list_length !pos_pions) < 3 then false
else let [(a,b);(c,d);(e,f)]= !pos_pions in
     a*(d-f) - b*(c-e) + c*f - d*e = 0
     (*c'est le dterminant de la matrice pour tester si les pions sont aligns*)
;;

let estime position =
let pos = position.plateau.pos in
let p = position.pions_restants in
let coeff = 5 + 10* (p.homme - p.machine) in (*25  85*)
if   test_gagn Homme pos
then if test_gagn Machine pos
     then coeff
     else mine
else if test_gagn Machine pos
     then maxe
else match pos.(1).(1) with
          |Rouge -> coeff + 2
          |Bleu  -> coeff - 2
          |Socle|Trou -> coeff
;;
(*doit toujours renvoyer entre mine et maxe,
le chiffre des dizaines correspond au nombre de pions poss.
*)

let inv_joueur = function
|Machine -> Homme
|Homme -> Machine
;;

let rec get_coup_machine profondeur joueur position =
let cp = ref (coups_possibles joueur position) in
let compare = ( match joueur with Machine -> (gt_int) |Homme -> (lt_int) ) in
let infini = ( match joueur with Machine -> mine; |Homme -> maxe ) in
     (*vir  la premire comparaison*)
let best_coup = ref {bcoup=hd !cp; valeur=infini} in
     (* modifier s'il n'y a pas toujours un coup possible*)
let switch = profondeur > 1 in
(try
  (while !cp <> [] do
     let a=hd !cp and t=tl !cp in
     cp:=t;
     let position_test = joue_coup a joueur position in
     if test_gagn joueur position_test.plateau.pos then
       begin best_coup:={ bcoup=a; valeur=(match joueur with Machine -> maxe |Homme -> mine) };
             raise Tangence;
       end;
     let coup__tester = if switch
                         then let val=(get_coup_machine (profondeur-1) (inv_joueur joueur)
                                                         position_test).valeur in
                                  {bcoup=a; valeur=val};
                         else {bcoup=a; valeur=estime position_test} in
     if compare coup__tester.valeur !best_coup.valeur then best_coup:=coup__tester;
  done)
with Tangence -> () );
!best_coup;
;;

let joue_coup_machine niveau position =
let coup__jouer = (get_coup_machine niveau Machine position).bcoup in
let result = joue_coup coup__jouer Machine position in
draw_plateau result.plateau;
result;
;;

(*enfin la conclusion*)
let main () =
clear_graph ();
let niveau = ref (choisir_niveau ()) in
let compteur = ref {homme=0; machine=0} in (*de type pions*)
clear_graph ();
while true do (*une partie*)
  let temp = qui_commence () in
  let position_courante = ref
    {plateau={case_vide=(1,1);
              pos=[|[|Socle;Socle;Socle|];
                    [|Socle;Trou ;Socle|];
                    [|Socle;Socle;Socle|]|]};
     coup_interdit=((0,0),(0,0));
     pions_restants={homme=3; machine=3}} in
  draw_plateau !position_courante.plateau;
  if temp then position_courante:=joue_coup_machine !niveau !position_courante;
  let pasfini = ref true in
  while !pasfini do (*un coup du joueur et de la machine*)
    let pasjou = ref true in
    while !pasjou do (*un coup du joueur*)
      try (position_courante:=joue_coup_homme !position_courante; pasjou:=false) with
      |Chgt Quitter -> raise (Chgt Quitter) (*c'est ici qu'il faut mettre exit 0*)
      |Chgt Niveau  -> (niveau:=choisir_niveau (); clear_graph ();
                        draw_plateau !position_courante.plateau)
      |Chgt Nouvelle_partie ->
         (pasfini:=false; pasjou:=false;
          compteur:={homme= !compteur.homme; machine= !compteur.machine +1})
    done;
    if test_gagn Homme !position_courante.plateau.pos
    then (pasfini:=false; compteur:={homme= !compteur.homme +1;
                                     machine= !compteur.machine})
    else begin position_courante:=joue_coup_machine !niveau !position_courante;
      if test_gagn Machine !position_courante.plateau.pos
      then (pasfini:=false; compteur:={homme= !compteur.homme;
                                       machine= !compteur.machine +1})
    
    end;
  done;
    print_pions !compteur;
done;;

main ();;
